from scipy import ndimage
from skimage.transform import radon
from scipy.signal import convolve
import numpy as np
import matplotlib.pyplot as plt
from cv2 import cv2
from scipy.fftpack import fft, ifft, fftfreq,fftshift
import os


#直接反投影,第一个参数为获得的正弦图,第二个参数为角度值,默认现在是180,因为得到正弦图时旋转了180个角度
def bp(sinogram, angels):
    #定义用于存储重建后的图像的数组
    channels = sinogram.shape[0]
    origin = np.zeros((angels, channels, channels))

    for i in range(angels):
    	#传入的图像中的每一列都对应于一个角度的投影值
        projectionValue = sinogram[:, i]
        #这里利用维度扩展和重复投影值数组来模拟反向均匀回抹过程

        #np.expand_dims是在指定位置插入新的轴来扩展数组形状,一维扩成二维
        projectionValueExpandDim = np.expand_dims(projectionValue, axis=0)

        #对数组中的元素进行连续重复复制,假设原来是1*N,则现在是channels*N
        projectionValueRepeat = projectionValueExpandDim.repeat(channels, axis=0)
        #获取指定的角度值，最后用于反向均匀回抹
        origin[i] = ndimage.rotate(projectionValueRepeat, i*180/angels, reshape=False).astype(np.float64)
    #各个投影角度的投影值已经都保存在origin数组中，只需要将它们相加即可
    image_bp = np.sum(origin, axis=0)
    return image_bp



#fbp
#第一个参数为获得的正弦图,第二个参数为角度值，默认现在是180，因为得到正弦图时旋转了180个角度
def fbp(sinogram, angels):
    #获得输出图像的大小
    channels = sinogram.shape[0]
    #定义用于存储重建后的图像的数组
    origin = np.zeros((angels, channels, channels))
    #为了得到ramp filter
    f = fftfreq(sinogram.shape[0]).reshape(-1, 1)   
    omega = 2 * np.pi * f                                
    
    # fourier_filter = 2 * np.abs(omega)
    #滤波器ramp filter的另一种实现方式
    a=0.1
    size,num_thetas=sinogram.shape
    step=2*np.pi/size
    w=np.arange(-np.pi,np.pi,step)
    if len(w)<size:
        w=np.concatenate([w,w[-1]+step])
    rn1=np.abs(2/a*np.sin(a*w/2))
    rn2=np.sin(a*w/2)/(a*w/2)
    r=rn1*(rn2)**2
    r = np.expand_dims(r, axis=1)                   
    fourier_filter=fftshift(r)
    #傅里叶变换并滤波
    projection = np.fft.fft(sinogram, axis=0) * fourier_filter
    # projection = np.fft.fft(sinogram, axis=0)
    #逆傅里叶变换
    proj_ifft = np.real(np.fft.ifft(projection, axis=0))
    # filter = RLFilter(channels, 1)

    for i in range(angels):
    	#传入的图像中的每一列都对应于一个角度的投影值
        projectionValue = proj_ifft[:, i]
        #这里利用维度扩展和重复投影值数组来模拟反向均匀回抹过程
        # projectionValueFiltered = convolve(filter, projectionValue, "same")
        projectionValueExpandDim = np.expand_dims(projectionValue, axis=0)
        #对数组中的元素进行连续重复复制,假设原来是1*N,则现在是channels*N
        projectionValueRepeat = projectionValueExpandDim.repeat(channels, axis=0)
        #获取指定的角度值，最后用于反向均匀回抹
        origin[i] = ndimage.rotate(projectionValueRepeat, i*180/angels, reshape=False).astype(np.float64)
    #各个投影角度的投影值已经都保存在origin数组中，只需要将它们相加即可
    image_fbp = np.sum(origin, axis=0)
    # image_fbp = (image_fbp*np.pi)/180
    return image_fbp


if __name__ == '__main__':
    # readPath = os.path.join(os.getcwd(),'111.png')
    readPath = 'C:/Users/qyf11/Desktop/444.png'
    #读取原始图片,以灰度方式读取
    image=cv2.imread(readPath,cv2.IMREAD_GRAYSCALE)
    # print(image.shape)
    #获得signorm图，默认投影角度为180°，所以得到的sinogram的第二个维度为180，第一个维度为原图的尺寸
    #如果需要改变角度，更改下面的180即可
    sinogram = radon(image,np.arange(180))
    #绘制原始图像和对应的sinogram图
    image_bp = bp(sinogram, sinogram.shape[1])
    image_fbp = fbp(sinogram, sinogram.shape[1])
    plt.subplot(2, 2, 1)
    plt.imshow(image, cmap='gray')
    plt.title('original image')

    plt.subplot(2, 2, 2)
    plt.imshow(sinogram, cmap='gray')
    plt.title('sinogram')

    plt.subplot(2, 2, 3)
    plt.imshow(image_bp, cmap='gray')
    plt.title('bp')

    plt.subplot(2, 2, 4)
    plt.imshow(image_fbp, cmap='gray')
    plt.title('fbp')
    #调整子图间距
    plt.subplots_adjust(wspace =0, hspace =1)
    plt.show()


